/**
 * @file
 * test.adjustColorBrightness node implementation.
 *
 * @copyright Copyright © 2012–2014 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "node.h"

#include "VuoImageRenderer.h"

VuoModuleMetadata({
                     "title" : "Kuwahara Filter",
                     "description" : "Kuwahara filter. 2014, George Toledo.",
					 "keywords" : [ ],
					 "version" : "1.0.0",
					 "dependencies" : [
						 "VuoImageRenderer"
					 ],
					 "node": {
						 "isInterface" : false
					 }
				 });

#include "node.h"

static const char * fragmentShaderSource = VUOSHADER_GLSL_SOURCE(120,
	// Inputs


	varying vec4 fragmentTextureCoordinate;

//2012, george toledo, based on GPUImage implementation
uniform sampler2D texture;
uniform int radius;
uniform float ResX;
uniform float ResY;

    void main (void)

{
    vec2 resolution=vec2(ResX,ResY);
    vec2 src_size = resolution;
    vec2 uv = fragmentTextureCoordinate.xy;
    float n = float((radius + 1) * (radius + 1));

vec3 m[4];
vec3 s[4];
    for (int k = 0; k < 4; ++k) {
    m[k] = vec3(0.0);
    s[k] = vec3(0.0);
    }

    for (int j = -radius; j <= 0; ++j) {
    for (int i = -radius; i <= 0; ++i) {
    vec3 c = texture2D(texture, uv + vec2(i,j) / src_size).rgb;
    m[0] += c;
    s[0] += c * c;
        }
    }

    for (int j = -radius; j <= 0; ++j) {
    for (int i = 0; i <= radius; ++i) {
    vec3 c = texture2D(texture, uv + vec2(i,j) / src_size).rgb;
    m[1] += c;
    s[1] += c * c;
        }
    }

    for (int j = 0; j <= radius; ++j) {
    for (int i = 0; i <= radius; ++i) {
    vec3 c = texture2D(texture, uv + vec2(i,j) / src_size).rgb;
    m[2] += c;
    s[2] += c * c;
        }
    }

    for (int j = 0; j <= radius; ++j) {
    for (int i = -radius; i <= 0; ++i) {
    vec3 c = texture2D(texture, uv + vec2(i,j) / src_size).rgb;
    m[3] += c;
    s[3] += c * c;
        }
    }


    float min_sigma2 = 1e+2;
    for (int k = 0; k < 4; ++k) {
    m[k] /= n;
    s[k] = abs(s[k] / n - m[k] * m[k]);

    float sigma2 = s[k].r + s[k].g + s[k].b;
    if (sigma2 < min_sigma2) {
    min_sigma2 = sigma2;
    gl_FragColor = vec4(m[k], 1.0);
        }
    }
}

);
struct nodeInstanceData
{
	VuoShader shader;
	VuoGlContext glContext;
	VuoImageRenderer imageRenderer;
};

struct nodeInstanceData * nodeInstanceInit(void)
{
	struct nodeInstanceData * instance = (struct nodeInstanceData *)malloc(sizeof(struct nodeInstanceData));
	VuoRegister(instance, free);

	instance->glContext = VuoGlContext_use();

    instance->shader = VuoShader_make("Kuwahara Filter");
	VuoShader_addSource(instance->shader, VuoMesh_IndividualTriangles, NULL, NULL, fragmentShaderSource);
	VuoRetain(instance->shader);

	instance->imageRenderer = VuoImageRenderer_make(instance->glContext);
	VuoRetain(instance->imageRenderer);

	return instance;
}

void nodeInstanceEvent
(
		VuoInstanceData(struct nodeInstanceData *) instance,
		VuoInputData(VuoImage) image,

        VuoInputData(VuoInteger, {"default":6,"suggestedStep":1}) radius,

		VuoOutputData(VuoImage) adjustedImage
)
{
	if (! image)
		return;

	// Feed parameters to the shader.
	VuoShader_setUniform_VuoImage((*instance)->shader, "texture", image);
    VuoShader_setUniform_VuoInteger ((*instance)->shader, "radius",  radius);

    VuoShader_setUniform_VuoReal ((*instance)->shader, "ResX",   image->pixelsWide);
    VuoShader_setUniform_VuoReal ((*instance)->shader, "ResY",   image->pixelsHigh);

	// Render.
	*adjustedImage = VuoImageRenderer_draw((*instance)->imageRenderer, (*instance)->shader, image->pixelsWide, image->pixelsHigh, VuoImage_getColorDepth(image));
}

void nodeInstanceFini(VuoInstanceData(struct nodeInstanceData *) instance)
{
	VuoRelease((*instance)->shader);
	VuoRelease((*instance)->imageRenderer);
	VuoGlContext_disuse((*instance)->glContext);
}
